//-----------------------------------------------------------------------------
// Copyright (c) 2007-2008 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This D3DX Effect file implements tangent space normal mapping.
//
// Three techniques are provided for each of the three lighting types:
// directional lighting, point lighting, and spotlight lighting.
//
// Tangent space normal mapping involves evaluating the lighting equations per
// pixel in the surface local coordinate space of the normal map texture.
// Rather than using the mesh's interpolated vertex normals tangent space
// normal mapping uses the normals from the normal map texture instead. Doing
// this allows additional surface detail to be applied to each face of the mesh
// without requiring extra geometry to be added to each face.
//
// The lighting vectors are transformed into the normal map texture's tangent
// space using a rotation matrix. This rotation matrix must be calculated for
// each face of the mesh. Each face has its own rotation matrix because each
// face is uniquely texture mapped to the normal map texture and the rotation
// matrix must take this into consideration. The rotation matrix is constructed
// using each face's tangent, bitangent, and normal vectors.
//
// The pixel shader used in tangent space normal mapping is identical to the
// pixel shader used in per-pixel lighting. The only difference is the normal
// used is read from the normal map texture rather that being interpolated by
// the rasterizer. Prior to being stored in the normal map texture the normals
// are in the range [-1,1]. Once stored in the normal map texture the normals
// will be compressed to the range [0,255]. When the normals are accessed from
// the pixel shader HLSL will remap then to the range [0,1]. To convert the
// normals back to the original [-1,1] range we need to perform a scale and a
// bias. We do this by: tex2D(normalMap, IN.texCoord) * 2.0f - 1.0f.
//
// Light attenuation for the point and spot lighting models is based on a
// light radius. Light is at its brightest at the center of the sphere defined
// by the light radius. There is no lighting at the edges of this sphere.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

//float4x4 g_mView;
float4x4 g_mWorld;
float4x4 g_mWorldInverseTranspose;
float4x4 g_mWorldViewProjection;
float4x4 g_mViewInv;

float4 blue = { 0.034f, 0.338f, 0.974f, 1.0f };
float4 green = { 0.05f, 0.30f, 0.15f, 1.0f };
float4 orange = {0.974f, 0.338f, 0.034f, 1.0f};
float4 grey = {0.12f, 0.103f, 0.092f, 1.0f};
float4 black = {0.0f, 0.0f, 0.0f, 1.0f};
float4 white = {1.0f, 1.0f, 1.0f, 1.0f};

float3 g_vLightDir;
float3 g_vLightPos;
float4 g_cLightDif;
float g_fLightRad;

bool g_outNormals;
bool g_mode;
bool g_bDefMat;
float g_fTiling = 5.0f;

float g_OverrideNormal = 1.0f;
float g_OverrideAlbedo = 1.0f;
float g_OverrideAO = 1.0f;
float g_fRoughness = 1.0f;
float g_fMetallic = 1.0f;
float g_LightIntensity  = 2.75f;//1.0f;
float g_fSpecular = 1.0f;
float g_AmbientLightIntensity = 1.0f;//0.5f;
#define PI 3.14159265359f

//-----------------------------------------------------------------------------
// Textures.
//-----------------------------------------------------------------------------

texture g_diffuseTexture;
texture g_normalTexture;
texture g_envTexture;
texture g_irradTexture;
texture g_rougTexture;
texture g_metallicTexture;
texture g_shadowTexture;
texture g_aoTexture;

sampler2D MetallicSampler = sampler_state
{
	Texture = <g_metallicTexture>;
	MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

sampler2D AOSampler = sampler_state
{
	Texture = <g_aoTexture>;
	MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

sampler2D ShadowSampler = sampler_state
{
	Texture = <g_shadowTexture>;
	MipFilter = Linear;
	MinFilter = Linear;
	MagFilter = Linear;
	AddressU = Clamp;
	AddressV = Clamp;
};

samplerCUBE irradMap = sampler_state
{
    Texture = <g_irradTexture>;
    MinFilter = Linear;
    MipFilter = Linear;
    MagFilter = Linear;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

samplerCUBE envMap = sampler_state
{
    Texture = <g_envTexture>;
    MinFilter = Linear;
    MipFilter = Linear;
    MagFilter = Linear;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

sampler2D roughnessMap = sampler_state
{
	Texture = <g_rougTexture>;
    MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

sampler2D colorMap = sampler_state
{
	Texture = <g_diffuseTexture>;
    MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

sampler2D normalMap = sampler_state
{
    Texture = <g_normalTexture>;
    MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

//-----------------------------------------------------------------------------
// Vertex Shaders.
//-----------------------------------------------------------------------------

struct VS_INPUT
{
	float3 position :  POSITION;
	float2 texCoord :  TEXCOORD0;
	float3 normal   :  NORMAL;
    float3 tangent  :  TANGENT;
};

struct VS_OUTPUT
{
	float4 position         :  POSITION;
	float2 texCoord         :  TEXCOORD0;
	float3 viewDir          :  TEXCOORD1;
	float3 WorldNormal      :  TEXCOORD2;
	float4 vScreenCoord	    :  TEXCOORD3;
	float3x3 tangentToWorld	:  TEXCOORD4;
};

VS_OUTPUT Standard_VS(VS_INPUT IN)
{
	VS_OUTPUT OUT;

	float3 worldPos = mul(float4(IN.position, 1.0f), g_mWorld).xyz;
	float3 n = normalize(mul(IN.normal, (float3x3)g_mWorld));
	float3 t = normalize(mul(IN.tangent, (float3x3)g_mWorld));
	float3 b = normalize(cross(n, t));
	float3x3 tbnMatrix = float3x3(t.x, b.x, n.x,
	                              t.y, b.y, n.y,
	                              t.z, b.z, n.z);

	OUT.position = mul(float4(IN.position, 1.0f), g_mWorldViewProjection);
	
	OUT.vScreenCoord.x = ( OUT.position.x * 0.5 + OUT.position.w * 0.5 );
	OUT.vScreenCoord.y = ( OUT.position.w * 0.5 - OUT.position.y * 0.5 );
	OUT.vScreenCoord.z = OUT.position.w;
	OUT.vScreenCoord.w = OUT.position.w;
	
	OUT.texCoord = IN.texCoord * g_fTiling;
	OUT.viewDir = g_mViewInv[3].xyz - worldPos;
	OUT.WorldNormal = n;
	OUT.tangentToWorld = transpose(tbnMatrix);

	return OUT;
}

//-----------------------------------------------------------------------------
// Pixel Shaders.
//-----------------------------------------------------------------------------
float NormalDistribution_GGX(float a, float NdH)
{
    // Isotropic ggx.
    float a2 = a*a;
    float NdH2 = NdH * NdH;

    float denominator = NdH2 * (a2 - 1.0f) + 1.0f;
    denominator *= denominator;
    denominator *= PI;

    return a2 / denominator;
}

float3 Fresnel_CookTorrance(float3 specularColor, float3 h, float3 v)
{
    float3 n = (1.0f + sqrt(specularColor)) / (1.0f - sqrt(specularColor));
    float c = saturate(dot(v, h));
    float3 g = sqrt(n * n + c * c - 1.0f);

    float3 part1 = (g - c)/(g + c);
    float3 part2 = ((g + c) * c - 1.0f)/((g - c) * c + 1.0f);

    return max(0.0f.xxx, 0.5f * part1 * part1 * ( 1 + part2 * part2));
}

float Geometric_CookTorrance(float a, float NdV, float NdL, float NdH, float VdH)
{
    return min(1.0f, min((2.0f * NdH * NdV)/VdH, (2.0f * NdH * NdL)/ VdH));
}

float3 Specular(float3 specularColor, float3 h, float3 v, float3 l, float a, float NdL, float NdV, float NdH, float VdH, float LdV)
{
    return ((NormalDistribution_GGX(a, NdH) * Geometric_CookTorrance(a, NdV, NdL, NdH, VdH)) * Fresnel_CookTorrance(specularColor, h, v) ) / (4.0f * NdL * NdV + 0.0001f);
}

float3 Diffuse(float3 pAlbedo)
{
    return pAlbedo/PI;
}

float3 ComputeLight(float3 albedoColor,float3 specularColor, float3 normal, float roughness, float3 lightPosition, float3 lightColor, float3 lightDir, float3 viewDir)
{
    // Compute some useful values.
    float NdL = saturate(dot(normal, lightDir));
    float NdV = saturate(dot(normal, viewDir));
    float3 h = normalize(lightDir + viewDir);
    float NdH = saturate(dot(normal, h));
    float VdH = saturate(dot(viewDir, h));
    float LdV = saturate(dot(lightDir, viewDir));
    float a = max(0.001f, roughness * roughness);

    float3 cDiff = Diffuse(albedoColor);
    float3 cSpec = Specular(specularColor, h, viewDir, lightDir, a, NdL, NdV, NdH, VdH, LdV);

    return lightColor * NdL * (cDiff * (1.0f - cSpec) + cSpec);
}

float4 Standard_PS(VS_OUTPUT IN) : COLOR
{
	float3 albedoColor = 0.0f;
	float3 normal = 0.0f;
	float roughness = 0.0f;
	float metallic = 0.0f;
	float reflection = 0.0f;
	float alpha = 0.0f;
	float overrideNormal = 0.0f;
	
	float fAO = 1.0f;
	
	if (g_bDefMat)
	{
	    float fCameraDethFade = 0.0f;
		overrideNormal = 0.3f;
	
	    float fBase = tex2D( colorMap, IN.texCoord * 0.5f ).g;
	    float fChecker = tex2D( colorMap, IN.texCoord * 5.0f ).g;
	    float fNoise = tex2D( colorMap, IN.texCoord * 20.0f ).r;
	    normal = normalize(2.0f * tex2D( normalMap, IN.texCoord * 10.0f ).rgb - 1.0f);
	    normal = normalize(mul(normal, IN.tangentToWorld));
		normal = lerp(normalize(IN.WorldNormal), normal, overrideNormal);
	
	    float fTempLerp = lerp(0.4f, 1.0f, fNoise);
		float fChecker1 = lerp(fChecker, 1.0f, fCameraDethFade);
	    fNoise = lerp(fTempLerp, 1.0f - fTempLerp, fChecker1);
		fChecker1 = lerp(fChecker, 0.0f, fCameraDethFade);
	    albedoColor = mul(fNoise, lerp(0.295f, 0.66f, lerp(fBase + fChecker1, 0.5f, 0.5f)));//mul(   , 0.5f)
		roughness = saturate(lerp(0.0f, 0.5f, fBase) + lerp(lerp(0.7f, 1.0f, fChecker), 1.0f, fCameraDethFade));
	
	    metallic = 0.0f;
	    reflection = 0.1f;
		alpha = 1.0f;
	}
	else
	{
	    float4 textureColor = tex2D( colorMap, IN.texCoord );
		alpha = textureColor.a;
		overrideNormal = g_OverrideNormal;
		fAO = lerp(1.0f, tex2D( AOSampler, IN.texCoord ).r, g_OverrideAO);
		
	    albedoColor = lerp(white, textureColor.rgb, g_OverrideAlbedo);
	    //albedoColor = pow(abs(albedoColor.rgb), 2.2f);
		normal = normalize(2.0f * tex2D(normalMap, IN.texCoord).rgb - 1.0f);
	    normal = normalize(mul(normal, IN.tangentToWorld));
		normal = lerp(normalize(IN.WorldNormal), normal, overrideNormal);
		
	    roughness = lerp(0.0f, 1.0f - tex2D( roughnessMap, IN.texCoord ).x, g_fRoughness);
	    metallic = lerp(1.0f, tex2D( MetallicSampler, IN.texCoord ).x, g_fMetallic);
	    reflection = g_fSpecular;
	}

	float3 lightDir = -normalize(g_vLightDir);
    float3 viewDir = normalize(IN.viewDir);
    
    float3 realAlbedo = albedoColor - albedoColor * metallic;
    float3 realSpecularColor = lerp(0.03f, albedoColor, metallic);
	
	float attenuation = 1.0f;
	float3 light1 = ComputeLight( realAlbedo, realSpecularColor,  normal,  roughness,  g_vLightPos, g_cLightDif, lightDir, viewDir);
	
	float3 cIrradColor = texCUBE( irradMap, normal);
	float3 reflectVector = reflect( viewDir, normal);
	float mipIndex =  roughness * roughness * 8.0f;
	float3 cEnvColor = texCUBElod( envMap, float4(reflectVector, mipIndex) );
	//cEnvColor = pow(abs(cEnvColor), 2.2f);
	//cIrradColor = pow(abs(cIrradColor), 2.2f);
	
	float3 envFresnel = Fresnel_CookTorrance(realSpecularColor, normal, viewDir);
	
	//float fShadowTerm = tex2Dproj( ShadowSampler, IN.vScreenCoord );

	float4 finalColor = 0.0f;
	if (g_outNormals)
	    finalColor = float4(normalize( IN.WorldNormal ) * 0.5 + 0.5, 1.0f);
	else
	{
	    if (g_mode)
	        finalColor = float4(attenuation * g_LightIntensity * light1 + envFresnel*cEnvColor * reflection + realAlbedo * cIrradColor * g_AmbientLightIntensity, alpha);
	    else
	        finalColor = float4( frac( IN.texCoord ), 0.0, 1.0 );//cIrradColor;
		finalColor.w = 1.0f;
	}
	
	return finalColor*fAO;// * fShadowTerm;

}

//-----------------------------------------------------------------------------
// Techniques.
//-----------------------------------------------------------------------------

technique Standard
{
	pass
	{
		VertexShader = compile vs_3_0 Standard_VS();
		PixelShader = compile ps_3_0 Standard_PS();
	}
}
